# Copyright (c) 2024 Sijmen Woutersen
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from bitarray import bitarray

from ..device import Pin, DiffPin, PinGroup

class HyperRAM:
    def __init__(self, CK: DiffPin, RESETn: Pin, CSn: Pin, RWDS: Pin, DQ: PinGroup):
        self.ctl = CK[0].device.ctl
        self.CK = CK
        self.RESETn = RESETn
        self.CSn = CSn
        self.RWDS = RWDS
        self.DQ = DQ

    async def init(self):
        self.CK.output_enable(True)
        self.RESETn.output_enable(True)
        self.CSn.output_enable(True)
        self.RWDS.output_enable(False)
        self.DQ.output_enable(False)
        self.CK.set_value(0)
        self.RESETn.set_value(0)
        self.CSn.set_value(1)
        await self.ctl.cycle()
        self.RESETn.set_value(1)
        await self.ctl.cycle()

    async def read(self, address, length=4, reg_space=False):
        try:
            self.DQ.output_enable(True)
            self.CSn.set_value(0)

            ca = [0x80 | (0x40 if reg_space else 0x00) | ((address >> 27) & 0x1f),
                (address >> 19) & 0xff,
                (address >> 11) & 0xff,
                (address >>  3) & 0xff,
                0,
                address & 0x07]

            for i in range(3):
                self.DQ.set_value(ca[2*i+0])
                await self.ctl.cycle()
                self.CK.set_value(1)
                await self.ctl.cycle()
                self.DQ.set_value(ca[2*i+1])
                await self.ctl.cycle()
                self.CK.set_value(0)
                await self.ctl.cycle()

            self.DQ.output_enable(False)

            r = []
            rwds_d = rwds_dd = 0
            for i in range((2*7+length//2)*4+50):
                self.CK.set_value(1 if i & 2 else 0)
                await self.ctl.cycle()

                if rwds_dd != rwds_d: r.append(self.DQ.get_value())
                rwds_dd = rwds_d
                rwds_d = self.RWDS.get_value()

                if len(r) == length: break

            if len(r) != length:
                raise Exception("HyperRAM not responding")
        finally:
            self.CSn.set_value(1)
            await self.ctl.cycle()
            self.DQ.output_enable(False)
            self.CK.set_value(0)
            await self.ctl.cycle()

        return r

    async def write(self, address, data, reg_space=False):
        try:
            self.DQ.output_enable(True)
            self.CSn.set_value(0)

            ca = [0x00 | (0x40 if reg_space else 0x00) | ((address >> 27) & 0x1f),
                (address >> 19) & 0xff,
                (address >> 11) & 0xff,
                (address >>  3) & 0xff,
                0,
                address & 0x07]

            for i in range(3):
                self.DQ.set_value(ca[2*i+0])
                await self.ctl.cycle()
                self.CK.set_value(1)
                await self.ctl.cycle()
                self.DQ.set_value(ca[2*i+1])
                await self.ctl.cycle()
                self.CK.set_value(0)
                await self.ctl.cycle()

            for i in range(2*7-1):
                await self.ctl.cycle()
                self.CK.set_value(1)
                await self.ctl.cycle()
                await self.ctl.cycle()
                self.CK.set_value(0)
                await self.ctl.cycle()

            self.RWDS.output_enable(True)
            self.RWDS.set_value(0)
            for i in range(len(data)//2):
                self.DQ.set_value(data[2*i])
                await self.ctl.cycle()
                self.CK.set_value(1)
                await self.ctl.cycle()
                self.DQ.set_value(data[2*i+1])
                await self.ctl.cycle()
                self.CK.set_value(0)
                await self.ctl.cycle()

        finally:
            self.CSn.set_value(1)
            await self.ctl.cycle()
            self.RWDS.output_enable(False)
            self.DQ.output_enable(False)
            await self.ctl.cycle()




class Crc:
    CCITT_CRC =  0x8408

    def __init__(self):
        self.crc = 0xffff

    def update(self, s):
        for in_byte in s:
            in_byte = ord(in_byte)
            if in_byte != 13:
                for _ in range(8):
                    feedback = (in_byte ^ self.crc) & 0x01;
                    self.crc >>= 1; # shift the shift register
                    if feedback: self.crc ^= Crc.CCITT_CRC; # invert selected bits
                    in_byte >>= 1; # get the next bit of in_byte

    def finalize(self):
        return (~self.crc) & 0xFFFF



# Copyright (c) 2025 Sijmen Woutersen
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import logging

from ..tap_controller import State
from .data import Variable, Evaluatable, Int, IntArray, Bool, BoolArray, CheckedVariableScope, ArrayVariable
from .stapl import (AssignmentInstruction, BooleanInstruction, CallInstruction, DataInstruction, DrScanInstruction,
                    DrStopInstruction, EndDataInstruction, EndProcedureInstruction, ExitInstruction, ExportInstruction,
                    ForInstruction, GotoInstruction, IfInstruction, IntegerInstruction, IrScanInstruction,
                    IrStopInstruction, NextInstruction, PopInstruction, PostDrInstruction, PostIrInstruction,
                    PreDrInstruction, PreIrInstruction, PrintInstruction, ProcedureInstruction, FrequencyInstruction,
                    PushInstruction, StateInstruction, TRSTInstruction, WaitInstruction)
from . import errors # type: ignore

logger = logging.getLogger(__name__)

class StaplExitCode(errors.StaplError):
    def __init__(self, code=0):
        self.code = code

class StaplInterpreter:
    class State:
        def __init__(self, pc, procedure):
            self.pc = pc
            self.procedure = procedure
            self.loop_stack = []
            self.scope = CheckedVariableScope()
            self.stack = []

    def __init__(self, ctl, stapl):
        self.stapl = stapl
        self.ctl = ctl
        self.call_stack = []
        self.data_scopes = {}
        self.state = None
        self.ir_stop = State.RUN_TEST_IDLE
        self.dr_stop = State.RUN_TEST_IDLE

    def _assign(self, variable, value):
        assert not self.state is None
        if not variable.last is None:
            assert not variable.first is None
            first = int(variable.first.evaluate(self.state.scope))
            last = int(variable.last.evaluate(self.state.scope))
            logger.debug("Setting %s[%d:%d] to %s", variable.name, first, last, value)
            self.state.scope[variable.name].assign(slice(first, last), value)
        elif not variable.first is None:
            first = int(variable.first.evaluate(self.state.scope))
            logger.debug("Setting %s[%d] to %s", variable.name, first, value)
            self.state.scope[variable.name].assign(first, value)
        else:
            logger.debug("Setting %s to %s", variable.name, value)
            self.state.scope[variable.name].assign(value)

    def execute(self, instruction=None):
        assert not self.state is None
        if instruction is None:
            assert self.state.pc < len(self.stapl.statements)
            instruction = self.stapl.statements[self.state.pc].instruction
            self.state.pc += 1

        logger.debug("%d: %s", instruction.line, instruction)

        if isinstance(instruction, AssignmentInstruction):
            self._assign(instruction.variable, instruction.value.evaluate(self.state.scope))
        elif isinstance(instruction, BooleanInstruction):
            if instruction.length is None:
                if instruction.value is None:
                    v = Bool(0)
                else:
                    v = instruction.value.evaluate(self.state.scope)
                var = Variable(v)
            else:
                if not instruction.value is None:
                    v = instruction.value.evaluate()
                else:
                    v = BoolArray([0] * instruction.length)
                var = ArrayVariable(v)
            self.state.scope[instruction.name] = var
            logger.debug("Setting %s to %s...", instruction.name, v)
        elif isinstance(instruction, CallInstruction):
            self.call_stack.append(self.state)
            self.state = StaplInterpreter.State(self.stapl.procedures[instruction.procedure], instruction.procedure)
        elif isinstance(instruction, DataInstruction):
            pass
        elif isinstance(instruction, DrScanInstruction):
            in_array = instruction.data_array.evaluate(self.state.scope)
            length = int(instruction.length.evaluate(self.state.scope))
            if len(in_array) < length:
                in_array.extend(length)
            elif len(in_array) > length:
                in_array = in_array[length-1:0]
            out_array = self.ctl.dr_scan(in_array.to_bitarray(), self.dr_stop)
            out_array = BoolArray(out_array)
            if not instruction.capture_array is None:
                self._assign(instruction.capture_array, out_array)
            if not instruction.compare_array is None:
                compare_array = instruction.compare_array.evaluate(self.state.scope)
                compare_mask_array = instruction.compare_mask_array.evaluate(self.state.scope)
                if out_array & compare_mask_array == compare_array & compare_mask_array:
                    self._assign(instruction.compare_result, Bool(1))
                else:
                    self._assign(instruction.compare_result, Bool(0))
        elif isinstance(instruction, DrStopInstruction):
            if instruction.state in (State.TEST_LOGIC_RESET, State.RUN_TEST_IDLE, State.PAUSE_IR, State.PAUSE_DR):
                self.dr_stop = instruction.state
            else:
                raise errors.InvalidState(instruction.state.name)
        elif isinstance(instruction, EndDataInstruction):
            return False
        elif isinstance(instruction, EndProcedureInstruction):
            if len(self.call_stack) > 0:
                self.state = self.call_stack.pop()
            else:
                return False
        elif isinstance(instruction, ExitInstruction):
            raise StaplExitCode(instruction.exit_code.evaluate(self.state.scope))
        elif isinstance(instruction, ExportInstruction):
            s = ""
            for part in instruction.parts:
                if isinstance(part, Evaluatable):
                    s += str(part.evaluate(self.state.scope))
                else:
                    s += str(part)
            logger.debug("EXPORT %s: %s", instruction.key, s)
            self.ctl.export(instruction.key, s)
        elif isinstance(instruction, ForInstruction):
            start = instruction.start.evaluate(self.state.scope)
            step = instruction.step.evaluate(self.state.scope)
            end = instruction.end.evaluate(self.state.scope)
            self.state.loop_stack.append((instruction.var, step, end, self.state.pc))
            self.state.scope[instruction.var] = var = Variable(start)
        elif isinstance(instruction, FrequencyInstruction):
            self.ctl.set_frequency(int(instruction.frequency.evaluate(self.state.scope)))
        elif isinstance(instruction, GotoInstruction):
            try:
                assert not self.state.procedure is None
                self.state.pc = self.stapl.labels[self.state.procedure][instruction.label]
            except KeyError:
                raise errors.LabelNotDefined(instruction.label)
        elif isinstance(instruction, IfInstruction):
            v = instruction.condition.evaluate(self.state.scope)
            if Bool(v):
                self.execute(instruction.instruction)
        elif isinstance(instruction, IntegerInstruction):
            if instruction.length is None:
                if instruction.value is None:
                    v = Int(0)
                else:
                    v = instruction.value.evaluate(self.state.scope)
                var = Variable(v)
            else:
                assert isinstance(instruction.value, IntArray)
                v = IntArray([0] * instruction.length)
                if not instruction.value is None:
                    for i in range(instruction.length):
                        v[i] = instruction.value[i].evaluate(self.state.scope)
                var = ArrayVariable(v)
            self.state.scope[instruction.name] = var
            logger.debug("Setting %s to %s...", instruction.name, v)
        elif isinstance(instruction, IrScanInstruction):
            in_array = instruction.data_array.evaluate(self.state.scope)
            length = int(instruction.length.evaluate(self.state.scope))
            if len(in_array) < length:
                raise errors.StaplError(f"Instruction array of size {len(in_array)} doesn't match length {instruction.length}")
            if len(in_array) != length:
                in_array = in_array[length-1:0]
            out_array = self.ctl.ir_scan(in_array.to_bitarray(), self.ir_stop)
            out_array = BoolArray(out_array)
            if not instruction.capture_array is None:
                self._assign(instruction.capture_array, out_array)
            if not instruction.compare_array is None:
                compare_array = instruction.compare_array.evaluate(self.state.scope)
                compare_mask_array = instruction.compare_mask_array.evaluate(self.state.scope)
                if out_array & compare_mask_array == compare_array & compare_mask_array:
                    self._assign(instruction.compare_result, Bool(1))
                else:
                    self._assign(instruction.compare_result, Bool(0))
        elif isinstance(instruction, IrStopInstruction):
            if instruction.state in (State.TEST_LOGIC_RESET, State.RUN_TEST_IDLE, State.PAUSE_IR, State.PAUSE_DR):
                self.ir_stop = instruction.state
            else:
                raise errors.InvalidState(instruction.state.name)
        elif isinstance(instruction, NextInstruction):
            if len(self.state.loop_stack) == 0:
                raise errors.StaplError("NEXT without FOR")
            var, step, end, first_pc = self.state.loop_stack[-1]
            if instruction.var != var:
                raise errors.StaplError(f"NEXT variable {instruction.var} doesn't match FOR variable {var}")
            var = self.state.scope[var]
            cur = var.evaluate()
            if (Int(step) > 0 and cur < end) or (Int(step) < 0 and cur > end):
                var.assign(cur + step)
                self.state.pc = first_pc
            else:
                self.state.loop_stack.pop()
        elif isinstance(instruction, PopInstruction):
            v = self.state.stack.pop()
            self._assign(instruction.variable, v)
        elif isinstance(instruction, PrintInstruction):
            s = ""
            for part in instruction.parts:
                if isinstance(part, Evaluatable):
                    s += str(part.evaluate(self.state.scope))
                else:
                    s += str(part)
            print(s)

        elif isinstance(instruction, ProcedureInstruction):
            assert len(self.state.scope) == 0
            for dep in instruction.uses:
                proc = self.stapl.procedures.get(dep)
                data = self.data_scopes.get(dep)
                if (proc is None) and (data is None):
                    raise errors.StaplError("Dependency {dep} not found for procedure {instruction.name}")
                elif not ((proc is None) or (data is None)):
                    raise errors.StaplError("Dependency {dep} is ambiguous for procedure {instruction.name}")
                if not data is None:
                    self.state.scope.update(data)
        elif isinstance(instruction, PushInstruction):
            self.state.stack.append(instruction.value.evaluate(self.state.scope))
        elif isinstance(instruction, StateInstruction):
            for state in instruction.states:
                self.ctl.enter_state(state)
        elif isinstance(instruction, TRSTInstruction):
            self.ctl.trst(cycles=int(instruction.wait_cycles.evaluate(self.state.scope)),
                          usec=int(instruction.wait_usec.evaluate(self.state.scope)))
        elif isinstance(instruction, WaitInstruction):
            if not instruction.wait_state is None:
                self.ctl.enter_state(instruction.wait_state)
            else:
                self.ctl.enter_state(State.RUN_TEST_IDLE)
            self.ctl.wait(cycles=int(instruction.wait_cycles.evaluate(self.state.scope)),
                          usec=int(instruction.wait_usec.evaluate(self.state.scope)))
            if not instruction.end_state is None:
                self.ctl.enter_state(instruction.end_state)
            else:
                self.ctl.enter_state(State.RUN_TEST_IDLE)
        else:
            raise NotImplementedError(f"{instruction} not implemented")

        return True

    def _run_procedure(self, pc, procedure=None):
        self.state = StaplInterpreter.State(pc, procedure)
        done = False
        while not done:
            try:
                done = not self.execute()
            except errors.StaplError as e:
                e.pc = pc
                raise
        state = self.state
        self.state = None
        return state

    def run(self, action, recommended=True, optional=False):
        self.ir_stop = State.RUN_TEST_IDLE
        self.dr_stop = State.RUN_TEST_IDLE

        which = ["required"]
        if recommended: which.append("recommended")
        if optional: which.append("optional")

        for name, pc in self.stapl.data_blocks.items():
            logger.debug("Initializing %s...", name)
            self.data_scopes[name] = self._run_procedure(pc).scope
            logger.debug("Data %s initialized", name)

        logger.info(f"Running action {action}...")
        try:
            action = self.stapl.actions[action]
        except KeyError:
            raise errors.StaplError(f"Action {action} not found") from None
        for procedure, opt in action.procedures:
            if opt in which:
                try:
                    pc = self.stapl.procedures[procedure]
                except KeyError:
                    raise errors.StaplError(f"Procedure {procedure} not found") from None
                else:
                    try:
                        self._run_procedure(pc, procedure)
                    except StaplExitCode as e:
                        if e.code == 0:
                            break
                        else:
                            raise
        logger.info(f"Action completed")




# Copyright (c) 2025 Sijmen Woutersen
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from ..device import Pin
from enum import Enum, auto

class BiasResistor(Enum):
    """Possible bias configurations for a digital line."""
    PULL_UP   = auto()   # Connect a resistor to VCC
    PULL_DOWN = auto()   # Connect a resistor to GND
    NONE      = auto()   # No bias resistor (floating)

class Net:
    """Class for testing nets between boundary scan cells

        name: net name for diagnostics
        ctls: list of controllers for all pins
        driver: driver pin (can be none for fixed pullled-up/down pins)
        receivers: receiver pin(s)
        bias: pull-up/down
    """
    def __init__(self, name: str, ctls, driver: Pin | None, receivers:Pin | list[Pin],
                 bias: BiasResistor=BiasResistor.NONE):
        self.name = name
        self.ctls = []
        if not driver is None:
            self.ctls.append(driver.device.ctl)
        self.driver = driver
        if isinstance(receivers, list):
            self.receivers = receivers
        else:
            self.receivers = [receivers]
        for receiver in self.receivers:
            self.ctls.append(receiver.device.ctl)
        self.bias = bias

    async def test(self):
        if not self.driver is None:
            self.driver.output_enable(True)
            for receiver in self.receivers:
                receiver.output_enable(False)
            for value in (0, 1, 0, 1):
                self.driver.set_value(value)
                for i in range(2):
                    for ctl in self.ctls:
                        await ctl.cycle()
                for receiver in self.receivers:
                    if receiver.get_value() != value:
                        raise Exception(f"{self.name} - {receiver} stuck at {0 if value else 1}")

        if self.bias != BiasResistor.NONE:
            if not self.driver is None:
                self.driver.output_enable(False)
            for receiver in self.receivers:
                receiver.output_enable(False)
            for i in range(2):
                for ctl in self.ctls:
                    await ctl.cycle()
            value = 1 if self.bias == BiasResistor.PULL_UP else 0
            if not self.driver is None and self.driver.get_value() != value:
                raise Exception(f"{self.name} - {self.driver} PULL-UP/DOWN not working")
            for receiver in self.receivers:
                if receiver.get_value() != value:
                    raise Exception(f"{self.name} - {receiver} PULL-UP/DOWN not working")



# Copyright (c) 2024 Sijmen Woutersen
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from bitarray import bitarray

class Driver:
    def reset(self):
        self.transmit_tms_str(bitarray('11111'))

    def set_freq(self, freq):
        pass

    def transfer(self, tms: int, tdi: int) -> int:
        raise NotImplementedError()

    def transmit_tms_str(self, tms_str: bitarray, tdi=0):
        for tms in tms_str:
            self.transfer(tms, tdi)

    def transfer_tdi_tdo_str(self, tdi_str: bitarray, first_tms=0, last_tms=0) -> bitarray:
        r = bitarray(endian='little')
        for tdi in tdi_str[:-1]:
            r.append(self.transfer(first_tms, tdi))
        r.append(self.transfer(last_tms, tdi_str[-1]))
        return r

    def transmit_tdi_str(self, tdi_str: bitarray, first_tms=0, last_tms=0):
        self.transfer_tdi_tdo_str(tdi_str, first_tms, last_tms)

    def receive_tdo_str(self, n, first_tms=0, first_tdi=0, last_tms=None, last_tdi=None) -> bitarray:
        if last_tms is None: last_tms = first_tms
        if last_tdi is None: last_tdi = first_tdi
        if n < 1: raise ValueError("n must be > 0")
        if n == 1 and first_tms != last_tms: raise ValueError("last_tms must be first_tms when n == 1")
        if n == 1 and first_tdi != last_tdi: raise ValueError("last_tdi must be first_tdi when n == 1")
        r = bitarray(endian='little')
        for i in range(n-1):
            r.append(self.transfer(first_tms, first_tdi))
        r.append(self.transfer(last_tms, last_tdi))
        return r



# Copyright (c) 2024 Sijmen Woutersen
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import asyncio

from bitarray import bitarray

from ..device import Pin, PinGroup, DiffPin

ON_FALLING = 0
BEFORE_RISING = 1
ON_RISING = 2
BEFORE_FALLING = 3

class DDR3:
    def __init__(self,
                 RESETn: Pin,
                 CK: DiffPin,
                 CKE: Pin | PinGroup,
                 CSn: Pin | PinGroup,
                 CASn: Pin,
                 RASn: Pin,
                 WEn: Pin,
                 A: PinGroup,
                 BA: PinGroup,
                 DQ: PinGroup,
                 DQS: DiffPin,
                 ODT: PinGroup,
                 DM: PinGroup):
        self.ctl = CK[0].device.ctl
        self.RESETn = RESETn
        self.CK = CK
        self.CKE = CKE
        self.CSn = CSn
        self.RASn = RASn
        self.CASn = CASn
        self.WEn = WEn
        self.A = A
        self.BA = BA
        self.DQ = DQ
        self.DQS = DQS
        self.ODT = ODT
        self.DM = DM
        self.cycle = None

    async def cycle_ck(self):
        self.CK.output_enable(True)
        while True:
            self.CK.set_value(0)
            self.cycle = ON_FALLING
            await self.ctl.cycle()
            self.cycle = BEFORE_RISING
            await self.ctl.cycle()
            self.CK.set_value(1)
            self.cycle = ON_RISING
            await self.ctl.cycle()
            self.cycle = BEFORE_FALLING
            await self.ctl.cycle()

    async def wait_cycle(self, cycle):
        while True:
            await self.ctl.cycle()
            if self.cycle == cycle: break

    async def before_rising_ck(self):
        await self.wait_cycle(BEFORE_RISING)

    async def after_rising_ck(self):
        await self.wait_cycle(BEFORE_FALLING)

    async def cmd_cycle(self, **pins):
        await self.before_rising_ck()

        for name, value in pins.items():
            o = getattr(self, name)
            if isinstance(o, list) or isinstance(o, tuple):
                for i in range(len(o)):
                    o[i].set_value(int(value[i]))
            else:
                o.set_value(int(value))

        await self.after_rising_ck()

        self.RASn.set_value(1)
        self.CASn.set_value(1)
        self.WEn.set_value(1)
        self.CSn.set_value(1)

    async def init(self):
        self.task = asyncio.create_task(self.cycle_ck())

        self.RESETn.output_enable(True)
        self.RESETn.set_value(0)
        self.CKE.output_enable(True)
        self.CKE.set_value(0)
        self.CSn.output_enable(True)
        self.CSn.set_value(1)
        self.DQ.output_enable(False)
        self.DQS.output_enable(False)
        self.A.output_enable(True)
        self.BA.output_enable(True)
        self.RASn.output_enable(True)
        self.RASn.set_value(1)
        self.CASn.output_enable(True)
        self.CASn.set_value(1)
        self.WEn.output_enable(True)
        self.WEn.set_value(1)
        self.ODT.output_enable(True)
        self.ODT.set_value(0)
        self.DM.output_enable(True)
        self.DM.set_value(0)

        await self.cmd_cycle()
        self.RESETn.set_value(1)
        await self.cmd_cycle()
        await self.cmd_cycle()
        await self.cmd_cycle()
        await self.cmd_cycle()
        await self.cmd_cycle()
        self.CKE.set_value(1)
        await self.cmd_cycle()

        # mr2
        mr2 = bitarray("0" * len(self.A))
        mr2[3:6] = bitarray("100") # CWL = 6
        await self.cmd_cycle(CSn="00", RASn=0, CASn=0, WEn=0, BA="010"[::-1], A=mr2)
        for i in range(4): await self.cmd_cycle()
        # mr3
        mr3 = bitarray("0" * len(self.A))
        await self.cmd_cycle(CSn="00", RASn=0, CASn=0, WEn=0, BA="011"[::-1], A=mr3)
        for i in range(4): await self.cmd_cycle()
        # mr1
        mr1 = bitarray("0" * len(self.A))
        mr1[0] = 1 # DLL = Disable
        mr1[3:5] = bitarray("00") # AL = 0
        await self.cmd_cycle(CSn="00", RASn=0, CASn=0, WEn=0, BA="001"[::-1], A=mr1)
        for i in range(4): await self.cmd_cycle()
        # mr0
        mr0 = bitarray("0" * len(self.A))
        mr0[0:2] = bitarray("00") # BL = Fixed BL8
        mr0[3] = 0 # BT = Sequential
        mr0[4:7] = bitarray("010") # CL = 6
        mr0[8] = 0 # DLL
        mr0[9:12] = bitarray("010") # WR = 6
        mr0[12] = 0 # PD
        await self.cmd_cycle(CSn="00", RASn=0, CASn=0, WEn=0, BA="000"[::-1], A=mr0)
        for i in range(12): await self.cmd_cycle()

    async def activate(self, ba=bitarray("000"), ra=bitarray("0000000000000000")):
        """Activate a row for reading or writing"""
        await self.cmd_cycle(CSn="00", RASn=0, BA=ba, A=ra)
        for i in range(6): await self.cmd_cycle()

    async def read(self, ba=bitarray("000"), ca=bitarray("0000000000000000")):
        """Read from active row"""
        r = []
        await self.cmd_cycle(CSn="00", CASn=0, BA=ba, A=ca)

        for i in range(20):
            if self.DQS.get_value() == 0: break
            await self.ctl.cycle()

        if self.DQS.get_value() == 1: raise Exception("Read timeout")

        for i in range(4):
            while self.DQS.get_value() == 0:
                await self.ctl.cycle()
            r.append(self.DQ.get_value())
            while self.DQS.get_value() == 1:
                await self.ctl.cycle()
            r.append(self.DQ.get_value())

        for i in range(6): await self.cmd_cycle()

        return r

    async def write(self, ba, ca, data, dm=0):
        """Write to active row"""
        await self.cmd_cycle(CSn="00", CASn=0, WEn=0, BA=ba, A=ca)
        for i in range(5):
            await self.cmd_cycle()

        self.DQ.output_enable(True)
        self.DQS.output_enable(True)
        self.DQS.set_value(0xFF)
        self.DM.set_value(dm)

        await self.ctl.cycle()

        for i in range(4):
            self.DQS.set_value(0)
            await self.ctl.cycle()
            self.DQ.set_value(data[2*i])
            await self.ctl.cycle()
            self.DQS.set_value(1)
            await self.ctl.cycle()
            self.DQ.set_value(data[2*i+1])
            await self.ctl.cycle()

        self.DQS.set_value(0)
        await self.ctl.cycle()

        self.DQ.output_enable(False)
        self.DQS.output_enable(False)

        for i in range(6): await self.cmd_cycle()

    async def precharge(self, ba=bitarray("000")):
        """Precharge active row (deactivating it)"""
        await self.cmd_cycle(CSn="00", RASn=0, WEn=0, BA=ba)
        for i in range(6): await self.cmd_cycle()



#!/usr/bin/env python3
import logging
import random

import ebyst

logger = logging.getLogger(__name__)

def test_fsm(count=10000):
    dev = ebyst.Device.from_bsdl("bsdl/BSDLLCMXO2-256HCQFN32.BSM")

    for state in ebyst.JtagState.__members__.values():
        sim = ebyst.drivers.Sim(dev)
        ctl = ebyst.TapController(sim)
        ctl._goto(state)
        if ctl.state != state or sim.state != state:
            logger.error(f"Incorrect state: target: {ebyst.JtagState(state).name}, controller: {ctl.state.name}, sim: {sim.state.name}")
            assert False

    sim = ebyst.drivers.Sim(dev)
    ctl = ebyst.TapController(sim)
    states = list(ebyst.JtagState.__members__.values())

    for i in range(count):
        state = states[random.randint(0, len(states)-1)]
        ctl._goto(state)
        if ctl.state != state or sim.state != state:
            logger.error(f"Incorrect state: target: {ebyst.JtagState(state).name}, controller: {ctl.state.name}, sim: {sim.state.name}")
            assert False

if __name__ == "__main__":
    logging.basicConfig()
    logging.getLogger().setLevel(logging.DEBUG)
    test_fsm()



# Copyright (c) 2024 Sijmen Woutersen
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import datetime

class Trace:
    def __init__(self, vcd_file, trace_all=False, **pins):
        self.pins = pins
        self.trace_all = trace_all
        self.t = 0
        self.ids = {}
        self.f = open(vcd_file, "wt")
        self.f.write(f"$version Generated by ebyst $end\n")
        self.f.write(f"$date {datetime.datetime.now()} $end\n")
        self.f.write(f"$timescale 1ps $end\n")
        self.f.write(f"$scope module TOP $end\n")
        for i, name in enumerate(pins):
            if isinstance(pins[name], list) or (isinstance(pins[name], tuple) and trace_all):
                n = len(pins[name])
            else:
                n = 1
            self.ids[name] = chr(i+65)
            self.f.write(f"$var wire {n} {chr(i+65)} {name} $end\n")
        self.f.write(f"$upscope $end\n")

    def __del__(self):
        self.f.close()

    def snapshot(self):
        self.f.write(f"#{self.t}\n")
        for name, pin in self.pins.items():
            if isinstance(pin, list) or (isinstance(pin, tuple) and self.trace_all):
                s = ""
                for pin in pin:
                    v = pin.get_value()
                    if v is None: v = 'X'
                    s = str(v) + s
                self.f.write(f"b{s} {self.ids[name]}\n")
            else:
                v = pin.get_value()
                if v is None: v = 'X'
                self.f.write(f"{v}{self.ids[name]}\n")
        self.t += 1


# Copyright (c) 2024 Sijmen Woutersen
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import pyparsing as pp
import re
from .data import Evaluatable, Int, Bool, BoolArray, Any, String, VariableScope
from . import aca, errors # type: ignore
from bitarray import bitarray
from bitarray.util import hex2ba, int2ba, ba2int

class VariableRef(Evaluatable):
    def __init__(self,  _s, _loc, tokens):
        if len(tokens) == 1:
            self.name = tokens[0]
            self.slice_start = None
            self.slice_end = None
        elif len(tokens) == 2:
            self.name = tokens[0]
            self.slice_start = tokens[1]
            self.slice_end = None
        elif len(tokens) == 3:
            self.name = tokens[0]
            self.slice_start = tokens[1]
            self.slice_end = tokens[2]
        else:
            print(tokens)
            assert False

    def evaluate(self, scope=VariableScope()):
        variable = scope[self.name]
        if not self.slice_end is None:
            assert not self.slice_start is None
            return variable.evaluate(scope)[slice(int(self.slice_start.evaluate(scope)), int(self.slice_end.evaluate(scope)))]
        elif not self.slice_start is None:
            return variable.evaluate(scope)[int(self.slice_start.evaluate(scope))].evaluate(scope)
        else:
            return variable.evaluate(scope)

    def __str__(self):
        if not self.slice_end is None:
            return f"{self.name}[{self.slice_start}..{self.slice_end}]"
        elif not self.slice_start is None:
            return f"{self.name}[{self.slice_start}]"
        else:
            return self.name

class BoolArrayParser(Evaluatable):
    def __init__(self, _s, _loc, tokens):
        assert len(tokens) == 1
        self.s = tokens[0]
        assert self.s[0] in "#@$"
        self.ba = None

    def evaluate(self, scope=VariableScope()):
        if self.ba is None:
            if self.s[0] == '#':
                self.ba = bitarray(re.sub(r'\s+', '', self.s[1:]), endian='little')
                self.ba.reverse()
            elif self.s[0] == '$':
                self.ba = bitarray(hex2ba(re.sub(r'\s+', '', self.s[1:]), endian='big'), 'little')
                self.ba.reverse()
            elif self.s[0] == '@':
                self.ba = bitarray(endian='little')
                self.ba.frombytes(aca.decompress(self.s[1:]))
            else:
                assert False

        return BoolArray(self.ba)

    def __str__(self):
        return self.s

class IntParser(Evaluatable):
    def __init__(self, _s, _loc, tokens):
        assert len(tokens) == 1
        self.v = int(tokens[0])

    def evaluate(self, scope=VariableScope()):
        if self.v == 0 or self.v == 1:
            return Any(self.v)
        else:
            return Int(self.v)

    def __str__(self):
        return str(self.v)

class Function(Evaluatable):
    def __init__(self,  _s, _loc, tokens):
        assert len(tokens) == 1
        assert len(tokens[0]) == 2
        self.function = tokens[0][0]
        self.v = tokens[0][1]

    def evaluate(self, scope={}):
        if self.function == "BOOL":
            ba = int2ba(int(self.v.evaluate(scope)), length=32, signed=True, endian='little')
            return BoolArray(ba)
        elif self.function == "INT":
            ba = self.v.evaluate(scope).v
            if len(ba) < 32:
                ba = ba + bitarray(32 - len(ba), endian='little')
            if len(ba) > 32:
                ba = ba[:32]
            return Int(ba2int(ba, signed=True))
        elif self.function == "CHR$":
            return String(chr(Int(self.v.evaluate(scope)).v))
        else:
            assert False

    def __str__(self):
        return f"{self.function}({self.v})"

class Expression(Evaluatable):
    def __init__(self,  _s, _loc, tokens):
        self.v = list(tokens)

    def optimize(self):
        try:
            return self.evaluate()
        except errors.VariableNotDefined:
            if len(self.v) == 1:
                return self.v[0].optimize()
            else:
                for i in range(len(self.v)):
                    try:
                        self.v[i] = self.v[i].optimize()
                    except AttributeError:
                        pass
                return self

    def evaluate(self, scope=VariableScope()):
        if len(self.v) == 1:
            return self.v[0].evaluate(scope)
        elif len(self.v) == 2:
            if self.v[0] == "~":
                return ~Int(self.v[1].evaluate(scope))
            elif self.v[0] == "-":
                return -Int(self.v[1].evaluate(scope))
            elif self.v[0] == "!":
                return ~Bool(self.v[1].evaluate(scope))
            else:
                print(self)
                assert False
        elif len(self.v) >= 3 and (len(self.v) & 1) == 1:
            r = self.v[0].evaluate(scope).clone()
            for i in range(1, len(self.v), 2):
                b = self.v[i+1].evaluate(scope)
                if self.v[i] == "*":
                    r *= b
                elif self.v[i] == "/":
                    r //= b
                elif self.v[i] == "%":
                    r %= b
                elif self.v[i] == "+":
                    r += b
                elif self.v[i] == "-":
                    r -= b
                elif self.v[i] == "<<":
                    r <<= b
                elif self.v[i] == ">>":
                    r >>= b
                elif self.v[i] == "&":
                    r &= b
                elif self.v[i] == "^":
                    r ^= b
                elif self.v[i] == "|":
                    r |= b
                elif self.v[i] == "<=":
                    r = r <= b
                elif self.v[i] == "<":
                    r = r < b
                elif self.v[i] == ">=":
                    r = r >= b
                elif self.v[i] == ">":
                    r = r > b
                elif self.v[i] == "==":
                    r = r == b
                elif self.v[i] == "!=":
                    r = r != b
                elif self.v[i] == "&&":
                    r = Bool(r)
                    r &= b
                elif self.v[i] == "||":
                    r = Bool(r)
                    r |= b
                else:
                    print(self)
                    assert False
            assert isinstance(r, Bool) or isinstance(r, Int) or isinstance(r, Any)
            return r
        else:
            print(self.v)
            assert False

    def __str__(self):
        return "(" + "".join([str(v) for v in self.v]) + ")"

    @classmethod
    def get_parse_rule(cls):
        expression = pp.Forward()
        variable = (pp.Regex(r"[a-zA-Z][a-zA-Z0-9_]*") +
                    pp.Opt(pp.Literal("[").suppress() + pp.Opt(expression + pp.Opt(pp.Literal("..").suppress() + expression)) +
                           pp.Literal("]").suppress())).set_parse_action(VariableRef)
        literal = pp.pyparsing_common.integer.set_parse_action(IntParser) | (pp.MatchFirst((
                                  pp.Regex(r"#[01\s]+"),
                                  pp.Regex(r"\$[0-9a-fA-F\s]+"),
                                  pp.Regex(r"@[^;]+")))).set_parse_action(BoolArrayParser)
        function = (pp.Group((pp.CaselessKeyword("BOOL") | pp.CaselessKeyword("INT") | pp.CaselessKeyword("CHR$")) +
                             pp.Literal("(").suppress() + expression + pp.Literal(")").suppress())).set_parse_action(Function)

        expression0 = (function | variable | literal).set_parse_action(cls)
        expression1 = expression0 | (pp.Literal("(").suppress() - expression - pp.Literal(")").suppress())
        expression2 = (pp.Opt(pp.one_of("- ! ~")) + expression1).set_parse_action(cls)
        expression3 = (expression2 - pp.ZeroOrMore(pp.one_of("* / %") + expression2)).set_parse_action(cls)
        expression4 = (expression3 - pp.ZeroOrMore(pp.one_of("+ -") + expression3)).set_parse_action(cls)
        expression5 = (expression4 - pp.ZeroOrMore(pp.one_of("<< >>") + expression4)).set_parse_action(cls)
        expression6 = (expression5 - pp.ZeroOrMore(pp.one_of("<= >= < >") + expression5)).set_parse_action(cls)
        expression7 = (expression6 - pp.ZeroOrMore(pp.one_of("== !=") + expression6)).set_parse_action(cls)
        expression8 = (expression7 - pp.ZeroOrMore(pp.Literal("&") + expression7)).set_parse_action(cls)
        expression9 = (expression8 - pp.ZeroOrMore(pp.Literal("^") + expression8)).set_parse_action(cls)
        expression10 = (expression9 - pp.ZeroOrMore(pp.Literal("|") + expression9)).set_parse_action(cls)
        expression11 = (expression10 - pp.ZeroOrMore(pp.Literal("&&") + expression10)).set_parse_action(cls)
        expression12 = (expression11 - pp.ZeroOrMore(pp.Literal("||") + expression11)).set_parse_action(cls)
        expression <<= expression12.set_parse_action(cls, lambda _s, _loc, tokens: tokens[0].optimize()) # type: ignore
        return expression



from .i2c import I2C
from .spi import SPI
from .mdio import MDIO
from .flash import MT25Q, W25Q
from .ddr3 import DDR3
from .ddr4 import DDR4
from .hyperram import HyperRAM
from .net import Net, BiasResistor



# Copyright (c) 2024 Sijmen Woutersen
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from ..device import Pin

class I2C:
    class NackError(Exception):
        def __init__(self):
            Exception.__init__(self, "I2C Nack")

    def __init__(self, SCL: Pin, SDA: Pin, address_bits=8, data_bits=8):
        if address_bits & 7 != 0: raise ValueError("address_bits must be a multiple of 8")
        if data_bits & 7 != 0: raise ValueError("address_bits must be a multiple of 8")
        self.ctl = SCL.device.ctl
        self.SCL = SCL
        self.SDA = SDA
        self.address_bits = address_bits
        self.data_bits = data_bits

    async def init(self):
        self.SCL.output_enable(True)
        self.SDA.output_enable(True)
        self.SCL.set_value(1)
        self.SDA.set_value(1)
        await self.ctl.cycle()

    async def _start(self):
        self.SCL.output_enable(True)
        self.SDA.output_enable(True)
        self.SDA.set_value(0)
        await self.ctl.cycle()
        self.SCL.set_value(0)
        await self.ctl.cycle()
    
    async def _clock_out_bit(self, bit: int):
        self.SCL.output_enable(True)
        self.SDA.output_enable(True)
        self.SDA.set_value(bit)
        await self.ctl.cycle()
        self.SCL.set_value(1)
        await self.ctl.cycle()
        self.SCL.set_value(0)
        await self.ctl.cycle()

    async def _clock_in_bit(self):
        self.SCL.output_enable(True)
        self.SDA.output_enable(False)
        await self.ctl.cycle()
        self.SCL.set_value(1)
        await self.ctl.cycle()
        r = self.SDA.get_value()
        self.SCL.set_value(0)
        await self.ctl.cycle()
        return r

    async def _restart(self):
        self.SCL.output_enable(True)
        self.SDA.output_enable(True)
        self.SDA.set_value(1)
        await self.ctl.cycle()
        self.SCL.set_value(1)
        await self.ctl.cycle()
        self.SDA.set_value(0)
        await self.ctl.cycle()
        self.SCL.set_value(0)
        await self.ctl.cycle()

    async def _stop(self):
        self.SCL.output_enable(True)
        self.SDA.output_enable(True)
        self.SCL.set_value(1)
        await self.ctl.cycle()
        self.SDA.set_value(1)
        await self.ctl.cycle()

    async def write(self, dev_address, reg_address=None, data=0):
        await self._start()

        for i in range(7):
            await self._clock_out_bit((dev_address >> (6-i)) & 1)
        await self._clock_out_bit(0)
        if await self._clock_in_bit(): raise I2C.NackError()

        for i in range(self.address_bits):
            await self._clock_out_bit((reg_address >> (self.address_bits-1-i)) & 1)
            if (i & 7) == 7:
                if await self._clock_in_bit(): raise I2C.NackError()

        for i in range(self.data_bits):
            await self._clock_out_bit((data >> (7-i)) & 1)
            if (i & 7) == 7:
                if await self._clock_in_bit(): raise I2C.NackError()

        await self._stop()

    async def read(self, dev_address, reg_address=None):
        d = 0
        await self._start()

        try:
            if self.address_bits > 0:
                if reg_address is None: raise ValueError("reg_address cannot be None with >0 address bits")
                for i in range(7):
                    await self._clock_out_bit((dev_address >> (6-i)) & 1)
                await self._clock_out_bit(0)
                if await self._clock_in_bit(): raise I2C.NackError()

                for i in range(self.address_bits):
                    await self._clock_out_bit((reg_address >> (self.address_bits-1-i)) & 1)
                    if i & 7 == 7:
                        if await self._clock_in_bit(): raise I2C.NackError()

                await self._restart()
            else:
                if not reg_address is None: raise ValueError("reg_address must be None with 0 address bits")

            for i in range(7):
                await self._clock_out_bit((dev_address >> (6-i)) & 1)
            await self._clock_out_bit(1)
            if await self._clock_in_bit(): raise I2C.NackError()

            for i in range(self.data_bits):
                d <<= 1
                d |= await self._clock_in_bit()
                if i == self.data_bits-1:
                    await self._clock_out_bit(1) # final bit; NACK
                elif i & 7 == 7:
                    await self._clock_out_bit(0) # ACK

        finally:
            await self._stop()

        return d



#!/usr/bin/env python3
import os, sys
import logging

from ebyst.stapl import StaplFile, StaplInterpreter, StaplExitCode
from ebyst import JtagState as State

class ColorFormatter(logging.Formatter):
    """Custom formatter to add colors to log levels."""

    # Define color codes
    COLORS = {
        "DEBUG": "\033[0;90m",
        "INFO": "\033[1;97m",
        "WARNING": "\033[1;33m",
        "ERROR": "\033[1;31m",
        "CRITICAL": "\033[1;41m"
    }

    RESET = "\033[0;0m"

    def format(self, record):
        log_color = self.COLORS.get(record.levelname, self.RESET)
        record.levelname = f"{log_color}{record.levelname}{self.RESET}"
        record.msg = f"{log_color}{record.msg}{self.RESET}"
        return super().format(record)

logger = logging.getLogger(__name__)

def get_all_stapls(root):
    ret = []
    for (root, dirs, files) in os.walk(root):
        for dir in dirs:
            ret += get_all_stapls(dir)
        for file in files:
            if os.path.splitext(file.lower())[-1] in (".stapl", ".stp"):
                ret.append(os.path.join(root, file))
    return ret

class Checker:
    def __init__(self, stapl):
        self.checks = []
        self.exit_code = None
        for note in stapl.notes:
            if note.key == "TEST":
                self.checks.append(str(note.text))
            elif note.key == "EXIT":
                self.exit_code = int(note.text)
        self.total = 0
        self.errors = 0

    def check(self, label):
        if len(self.checks) == 0:
            logger.error(f"CHECK {label} == ??: FAIL")
            self.errors += 1
        else:
            if self.checks[0] == str(label):
                logger.debug(f"CHECK {label} == {self.checks[0]}: OK")
            else:
                logger.error(f"CHECK {label} == {self.checks[0]}: FAIL")
                self.errors += 1
            self.checks.pop(0)
        self.total += 1

    def export(self, key, data):
        if key != "TEST": return
        self.check(data)

    def trst(self, cycles, usec):
        self.check(f"TRST {cycles} CYCLES, {usec} USEC")

    def enter_state(self, state):
        if state == State.TEST_LOGIC_RESET:
            self.check(f"STATE RESET")
        elif state == State.RUN_TEST_IDLE:
            self.check(f"STATE IDLE")
        elif state == State.SELECT_DR_SCAN:
            self.check(f"STATE DRSELECT")
        elif state == State.CAPTURE_DR:
            self.check(f"STATE DRCAPTURE")
        elif state == State.SHIFT_DR:
            self.check(f"STATE DRSHIFt")
        elif state == State.EXIT1_DR:
            self.check(f"STATE DREXIT1")
        elif state == State.PAUSE_DR:
            self.check(f"STATE DRPAUSE")
        elif state == State.EXIT2_DR:
            self.check(f"STATE DREXIT2")
        elif state == State.UPDATE_DR:
            self.check(f"STATE DRUPDATE")
        elif state == State.SELECT_IR_SCAN:
            self.check(f"STATE IRSELECT")
        elif state == State.CAPTURE_IR:
            self.check(f"STATE IRCAPTURE")
        elif state == State.SHIFT_IR:
            self.check(f"STATE IRSHIFT")
        elif state == State.EXIT1_IR:
            self.check(f"STATE IREXIT1")
        elif state == State.PAUSE_IR:
            self.check(f"STATE IRPAUSE")
        elif state == State.EXIT2_IR:
            self.check(f"STATE IREXIT2")
        elif state == State.UPDATE_IR:
            self.check(f"STATE IRUPDATE")
        else:
            print(state)
            assert False

    def ir_scan(self, ir, end_state):
        self.check(f"IR SCAN {ir.to01()[::-1]}")
        self.enter_state(end_state)
        return ir

    def dr_scan(self, dr, end_state):
        self.check(f"DR SCAN {dr.to01()[::-1]}")
        self.enter_state(end_state)
        return dr

    def wait(self, cycles, usec):
        self.check(f"WAIT {cycles} CYCLES, {usec} USEC")

    def is_done(self):
        return len(self.checks) == 0

def test_action(stapl_file, action):
    chk = Checker(stapl_file)
    interpreter = StaplInterpreter(chk, stapl_file)
    logger.info(f"Running {action}")
    exit_code = None
    try:
        interpreter.run(action)
    except StaplExitCode as e:
        exit_code = e.code
    if not chk.is_done():
        raise Exception(f"EXPORT TEST output count doesn't match NOTE TEST output count")
    if chk.exit_code != exit_code:
        raise Exception(f"EXIT code doesn't match NOTE EXIT output")
    if chk.total == 0:
        raise Exception(f"No checks found")
    if chk.errors > 0:
        raise Exception(f"One or more checks failed")

if __name__ == "__main__":
    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)
    formatter = ColorFormatter("[%(asctime)s] [%(levelname)s] %(message)s", "%H:%M:%S")
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setFormatter(formatter)
    logger.addHandler(console_handler)

    if len(sys.argv) == 1:
        for fn in sorted(get_all_stapls("stapl/tests")):
            logger.info(f"Parsing {fn}")
            with open(fn, "r") as f:
                stapl = StaplFile.parse(f)
            for action in stapl.actions:
                test_action(stapl, action)
    elif len(sys.argv) == 2:
        fn = sys.argv[1]
        logger.info(f"Parsing {fn}")
        with open(fn, "r") as f:
            stapl = StaplFile.parse(f)
        for action in stapl.actions:
            test_action(stapl, action)
    elif len(sys.argv) == 3:
        fn = sys.argv[1]
        logger.info(f"Parsing {fn}")
        with open(fn, "r") as f:
            stapl = StaplFile.parse(f)
        test_action(stapl, sys.argv[2])
    else:
        assert False


# Copyright (c) 2024 Sijmen Woutersen
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import asyncio

from bitarray import bitarray

from ..device import Pin, PinGroup, DiffPin

# MT0 = XOR (A1, A6, PAR)
# MT1 = XOR (A8, ALERT_n, A9)
# MT2 = XOR (A2, A5, A13)
# MT3 = XOR (A0 A7, A11)
# MT4 = XOR (CK_c, ODT, CAS_n/A15)
# MT5 = XOR (CKE, RAS_n,/A16, A10/AP)
# MT6 = XOR (ACT_n, A4, BA1)
# MT7 = XOR (BG1, DM, CK_t)
# MT8 = XOR (WE_n / A14, A12 / BC, BA0)
# MT9 = XOR (BG0, A3, (RESET_n and TEN))


# DQ0 = MT0
# DQ1 = MT1
# DQ2 = MT2
# DQ3 = MT3
# DQ4 = MT4
# DQ5 = MT5
# DQ6 = MT6
# DQ7 = MT7
# DQS_t = MT8
# DQS_c = MT9


class DDR4:
    def __init__(self,
                 RESETn: Pin,
                 TEN: Pin,
                 CK: DiffPin,
                 CKE: Pin | PinGroup,
                 CSn: Pin | PinGroup,
                 CASn: Pin,
                 RASn: Pin,
                 WEn: Pin,
                 BG: PinGroup,
                 A: PinGroup,
                 BA: PinGroup,
                 DQ: PinGroup,
                 DQS: DiffPin,
                 ODT: PinGroup,
                 ACTn: Pin,
                 PARITY: Pin,
                 ALERTn: Pin,
                 DM: Pin):
        self.ctl = CK[0].device.ctl
        self.RESETn = RESETn
        self.CK = CK
        self.CKE = CKE
        self.CSn = CSn
        self.RASn = RASn
        self.CASn = CASn
        self.WEn = WEn
        self.BG = BG
        self.A = A
        self.BA = BA
        self.DQ = DQ
        self.DQS = DQS
        self.ODT = ODT
        self.DM = DM
        self.ACTn = ACTn
        self.PARITY = PARITY
        self.ALERTn = ALERTn
        self.TEN = TEN

        # Connectivity test connections
        self.MT = {
            self.DQ[0]: (self.A[1], self.A[6], self.PARITY),
            self.DQ[1]: (self.A[8], self.ALERTn, self.A[9]),
            self.DQ[2]: (self.A[2], self.A[5], self.A[13]),
            self.DQ[3]: (self.A[0], self.A[7], self.A[11]),
            self.DQ[4]: (self.CK[1], self.ODT, self.CASn),
            self.DQ[5]: (self.CKE, self.RASn, self.A[10]),
            self.DQ[6]: (self.ACTn, self.A[4], self.BA[1]),
            self.DQ[7]: (self.BG[1], self.DM, self.CK[0]),
            self.DQS[0]: (self.WEn, self.A[12], self.BA[0]),
            self.DQS[1]: (self.BG[0], self.A[3], self.RESETn),
        }

    async def cycle(self, count=1):
        for i in range(count):
            self.CK.set_value(1)
            await self.ctl.cycle()
            self.CK.set_value(0)
            await self.ctl.cycle()

    async def init(self):
        self.CK.output_enable(True)
        self.CK.set_value(0)
        self.RESETn.output_enable(True)
        self.RESETn.set_value(0)
        self.CKE.output_enable(True)
        self.CKE.set_value(0)
        self.CSn.output_enable(True)
        self.CSn.set_value(3)
        self.DQ.output_enable(False)
        self.DQS.output_enable(False)
        self.A.output_enable(True)
        self.BA.output_enable(True)
        self.BG.output_enable(True)
        self.RASn.output_enable(True)
        self.RASn.set_value(1)
        self.CASn.output_enable(True)
        self.CASn.set_value(1)
        self.ACTn.output_enable(True)
        self.ACTn.set_value(1)
        self.ALERTn.output_enable(True)
        self.ALERTn.set_value(1)
        self.PARITY.output_enable(True)
        self.PARITY.set_value(0)
        self.WEn.output_enable(True)
        self.WEn.set_value(1)
        self.ODT.output_enable(True)
        self.ODT.set_value(0)
        self.DM.output_enable(True)
        self.DM.set_value(0)
        self.TEN.output_enable(True)
        self.TEN.set_value(1)
        self.RESETn.set_value(0)
        await self.ctl.cycle()
        self.RESETn.set_value(1)

    async def test(self):
        self.CSn.set_value(0)

        # all zero
        for v in self.MT.values():
            for i in range(3):
                v[i].set_value(0)
        await self.ctl.cycle()

        for v in self.MT.values():
            v[0].set_value(1)
        await self.ctl.cycle()
        for v in self.MT.keys():
            if v.get_value() != 0:
                raise Exception(f"Connectivity test failed on {self.MT[v][0].name} / {self.MT[v][1].name} / {self.MT[v][2].name} => {v.name}")

        for v in self.MT.values():
            v[1].set_value(1)
        await self.ctl.cycle()
        for v in self.MT.keys():
            if v.get_value() != 1:
                raise Exception(f"Connectivity test failed on {self.MT[v][0].name} => {v.name}")

        for v in self.MT.values():
            v[2].set_value(1)
        await self.ctl.cycle()
        for v in self.MT.keys():
            if v.get_value() != 0:
                raise Exception(f"Connectivity test failed on {self.MT[v][1].name} => {v.name}")

        await self.ctl.cycle()
        for v in self.MT.keys():
            if v.get_value() != 1:
                raise Exception(f"Connectivity test failed on {self.MT[v][2].name} => {v.name}")

        # TODO test for signals stuck together


#!/usr/bin/env python3
import asyncio
import logging

import ebyst

logger = logging.getLogger(__name__)

async def loopback_test(ctl, opin, ipin, count):
    opin.output_enable(True)
    ipin.output_enable(False)

    for i in range(count):
        if i > 1: assert i & 1 == ipin.get_value()
        opin.set_value(i & 1)
        await ctl.cycle()

async def main():
    logging.basicConfig()
    logging.getLogger().setLevel(logging.INFO)
    drv = ebyst.drivers.MPSSE(ebyst.drivers.MPSSE.list_devices()[0])
    dev = ebyst.Device.from_bsdl("bsdl/BSDLLCMXO2-256HCQFN32.BSM")
    ctl = ebyst.TapController(drv)
    ctl.detect_chain()
    ctl.add_device(dev)
    ctl.validate_chain()
    ctl.extest()
    async with asyncio.TaskGroup() as tg:
        tg.create_task(loopback_test(ctl, dev.pinmap['PB2C'], dev.pinmap['PB2A'], 100))
        tg.create_task(loopback_test(ctl, dev.pinmap['PB4C'], dev.pinmap['PB4D'], 70))
    assert ctl.cycle_counter == 100

if __name__ == "__main__":
    asyncio.run(main())


#!/usr/bin/env python3
#
# Control LEDs of MPF300TS eval kit via FlashPro Express
#
# Flashpro tcl script (execute with 'FPExpress SCRIPT:flashpro.tcl');
#
# file delete -force tmp
# new_project -name fp -location tmp -mode single
# set_programming_file -file {flashpro.stp}
# set_programming_action -action {BSCAN}
# run_selected_actions
# close_project

import logging
import asyncio
import sys

from bitarray import bitarray
from bitarray.util import int2ba, ba2int

import ebyst

from ebyst.interfaces import MT25Q, MDIO, DDR3, DDR4
from ebyst import Pin, PinGroup, DiffPin

logger = logging.getLogger(__name__)

async def leds(dev):
    LEDS_ON = ['IO_H21', 'IO_H22', 'IO_F23', 'IO_C27']
    LEDS_OFF = ['IO_D25', 'IO_C26', 'IO_B26', 'IO_F22']
    for pin in LEDS_ON:
        dev.pinmap[pin].output_enable()
        dev.pinmap[pin].set_value(True)
    for pin in LEDS_OFF:
        dev.pinmap[pin].output_enable()
        dev.pinmap[pin].set_value(False)
    await dev.ctl.cycle()
    await asyncio.sleep(2.0)

async def main():
    drv = ebyst.drivers.MPSSE(ebyst.drivers.MPSSE.list_devices([(0x1514, 0x2008)])[0])
    dev = ebyst.Device.from_bsdl("bsdl/MPF300TSFCG1152.bsdl")
    ctl = ebyst.TapController(drv)

    ctl.detect_chain()
    ctl.add_device(dev)
    ctl.validate_chain()

    ctl.start_stapl_recording(sys.stdout)

    try:
        ctl.extest()
        async with asyncio.TaskGroup() as tg:
            tg.create_task(leds(dev))

    except KeyboardInterrupt:
        pass
    finally:
        ctl.stop_stapl_recording(reset=False)
        ctl.reset()

if __name__ == "__main__":
    logging.basicConfig()
    logging.getLogger().setLevel(logging.INFO)
    asyncio.run(main())


#!/usr/bin/env python3
import unittest
from ebyst.stapl.data import Int, Bool, Any, IntArray, BoolArray, Variable, ArrayVariable
from ebyst.stapl import errors

class TestCalc(unittest.TestCase):
    def test_bool(self):
        self.assertTrue(Bool(True))
        self.assertFalse(Bool(False))

        self.assertTrue(Bool(True) == Bool(True))
        self.assertFalse(Bool(True) == Bool(False))
        self.assertFalse(Bool(False) == Bool(True))
        self.assertTrue(Bool(False) == Bool(False))
        self.assertIsInstance(Bool(True) == Bool(True), Bool)

        self.assertFalse(Bool(True) != Bool(True))
        self.assertTrue(Bool(True) != Bool(False))
        self.assertTrue(Bool(False) != Bool(True))
        self.assertFalse(Bool(False) != Bool(False))
        self.assertIsInstance(Bool(True) != Bool(True), Bool)

        self.assertTrue(Bool(True) and Bool(True))
        self.assertFalse(Bool(True) and Bool(False))
        self.assertFalse(Bool(False) and Bool(True))
        self.assertFalse(Bool(False) and Bool(False))
        self.assertIsInstance(Bool(True) and Bool(True), Bool)

        self.assertTrue(Bool(True) or Bool(True))
        self.assertTrue(Bool(True) or Bool(False))
        self.assertTrue(Bool(False) or Bool(True))
        self.assertFalse(Bool(False) or Bool(False))
        self.assertIsInstance(Bool(True) or Bool(True), Bool)

        self.assertFalse(not Bool(True))
        self.assertTrue(not Bool(False))
        self.assertFalse(~Bool(True))
        self.assertTrue(~Bool(False))
        self.assertIsInstance(~Bool(True), Bool)

        a = Bool(1)
        a &= 0
        self.assertEqual(a, 0)
        a |= 1
        self.assertEqual(a, 1)
        a ^= 1
        self.assertEqual(a, 0)
        self.assertIsInstance(a, Bool)

    def test_bool_var(self):
        a = Variable(Bool(True))
        self.assertEqual(a.evaluate(), Bool(True))
        a.assign(Bool(False))
        self.assertEqual(a.evaluate(), Bool(False))
        self.assertRaises(errors.StaplValueError, lambda: a.assign(Int(4)))

    def test_int(self):
        self.assertEqual(Int(1), Int(1))
        self.assertIsInstance(Int(1) == Int(1), Bool)
        self.assertNotEqual(Int(1), Int(2))
        self.assertIsInstance(Int(1) != Int(2), Bool)

        self.assertEqual(Int(1), 1)
        self.assertNotEqual(Int(1), 2)

        a = Int(1)
        self.assertEqual(a, 1)
        a += 4
        self.assertEqual(a, 5)
        a *= 2
        self.assertEqual(a, 10)
        a //= 3
        self.assertEqual(a, 3)
        a -= 1
        self.assertEqual(a, 2)
        a += 8
        self.assertEqual(a, 10)
        a %= 4
        self.assertEqual(a, 2)
        a <<= 3
        self.assertEqual(a, 16)
        a >>= 1
        self.assertEqual(a, 8)
        self.assertIsInstance(a, Int)

        self.assertEqual(-Int(4), -4)
        self.assertEqual(~Int(4), ~4)

        b = Int(0xaaaa)
        b |= Int(0x55)
        self.assertEqual(b, 0xaaff)
        b &= Int(0xa5ff)
        self.assertEqual(b, 0xa0ff)
        b ^= Int(0x0050)
        self.assertEqual(b, 0xa0af)
        self.assertIsInstance(a, Int)

        self.assertTrue(Int(4) > Int(1))
        self.assertFalse(Int(4) > Int(4))
        self.assertFalse(Int(1) > Int(4))
        self.assertIsInstance(Int(4) > Int(1), Bool)

        self.assertTrue(Int(4) >= Int(1))
        self.assertTrue(Int(4) >= Int(4))
        self.assertFalse(Int(1) >= Int(4))
        self.assertIsInstance(Int(4) >= Int(1), Bool)

        self.assertFalse(Int(4) < Int(1))
        self.assertFalse(Int(4) < Int(4))
        self.assertTrue(Int(1) < Int(4))
        self.assertIsInstance(Int(4) < Int(1), Bool)

        self.assertFalse(Int(4) <= Int(1))
        self.assertTrue(Int(4) <= Int(4))
        self.assertTrue(Int(1) <= Int(4))
        self.assertIsInstance(Int(4) <= Int(1), Bool)

        self.assertRaises(errors.StaplValueError, lambda: Int(True))
        self.assertRaises(errors.StaplValueError, lambda: Int(Bool(True)))
        self.assertRaises(errors.StaplValueError, lambda: Int(False))
        self.assertRaises(errors.StaplValueError, lambda: Int(Bool(False)))

    def test_int_var(self):
        a = Variable(Int(1))
        self.assertEqual(a.evaluate(), Int(1))
        a.assign(Int(3))
        self.assertEqual(a.evaluate(), Int(3))
        self.assertRaises(errors.StaplValueError, lambda: a.assign(Bool(True)))

    def test_any(self):
        self.assertEqual(Any(1), Any(1))
        self.assertIsInstance(Any(1) == Any(1), Bool)
        self.assertNotEqual(Any(1), Any(2))
        self.assertIsInstance(Any(1) != Any(2), Bool)

        self.assertEqual(Any(1), 1)
        self.assertNotEqual(Any(1), 2)

        a = Any(1)
        a += Int(1)
        self.assertEqual(a, 2)
        self.assertIsInstance(a, Int)

        a = Any(0)
        self.assertIsInstance(Bool(a), Bool)
        self.assertIsInstance(Int(a), Int)

        a = Any(1)
        self.assertIsInstance(Bool(a), Bool)
        self.assertIsInstance(Int(a), Int)

        a = Any(2)
        self.assertRaises(errors.StaplValueError, lambda: Bool(a))
        self.assertIsInstance(Int(a), Int)

        self.assertEqual(Any(1), Bool(1))
        self.assertEqual(Any(2), Int(2))
        self.assertNotEqual(Any(1), Bool(0))
        self.assertNotEqual(Any(1), Int(0))

        self.assertEqual(Bool(1), Any(1))
        self.assertEqual(Int(2), Any(2))
        self.assertNotEqual(Bool(1), Any(0))
        self.assertNotEqual(Int(1), Any(0))
        self.assertRaises(errors.StaplValueError, lambda: Bool(1) == Any(2))

    def test_int_array(self):
        a = IntArray([1, 2, 3, 4])
        self.assertEqual(len(a), 4)
        self.assertEqual(a[0], Int(1))
        self.assertEqual(a[1], Int(2))
        self.assertEqual(a[2], Int(3))
        self.assertEqual(a[3], Int(4))
        self.assertIsInstance(a[0], Int)

        b = a[1:2]
        self.assertIsInstance(b, IntArray)
        assert isinstance(b, IntArray)
        self.assertEqual(len(b), 2)
        self.assertEqual(b[0], Int(2))
        self.assertEqual(b[1], Int(3))

        c = a[3:1]
        self.assertIsInstance(c, IntArray)
        assert isinstance(c, IntArray)
        self.assertEqual(len(c), 3)
        self.assertEqual(c[0], Int(4))
        self.assertEqual(c[1], Int(3))
        self.assertEqual(c[2], Int(2))

    def test_int_array_var(self):
        a = ArrayVariable(IntArray([1, 2, 3, 4]))
        self.assertEqual(a.evaluate(), IntArray([1, 2, 3, 4]))
        self.assertRaises(errors.StaplValueError, lambda: a.assign(BoolArray("110011")))
        a.assign(IntArray([5, 6, 7, 8]))
        self.assertEqual(a.evaluate(), IntArray([5, 6, 7, 8]))
        a.assign(slice(1, 2), IntArray([9, 10]))
        self.assertEqual(a.evaluate(), IntArray([5, 9, 10, 8]))
        a.assign(3, Int(11))
        self.assertEqual(a.evaluate(), IntArray([5, 9, 10, 11]))
        a.assign(slice(2, 0), IntArray([12, 13, 14]))
        self.assertEqual(a.evaluate(), IntArray([14, 13, 12, 11]))
        self.assertRaises(errors.StaplValueError, lambda: a.assign(slice(2, 0), IntArray([15])))
        self.assertRaises(errors.StaplValueError, lambda: a.assign(1, IntArray([15])))
        self.assertRaises(errors.StaplValueError, lambda: a.assign(slice(2, 0), Int(15)))

    def test_bool_array(self):
        a = BoolArray("101011")
        self.assertEqual(len(a), 6)
        self.assertEqual(a[5], Bool(1))
        self.assertEqual(a[4], Bool(0))
        self.assertEqual(a[3], Bool(1))
        self.assertEqual(a[2], Bool(0))
        self.assertEqual(a[1], Bool(1))
        self.assertEqual(a[0], Bool(1))
        self.assertIsInstance(a[0], Bool)

        b = a[2:1]
        self.assertIsInstance(b, BoolArray)
        assert isinstance(b, BoolArray)
        self.assertEqual(len(b), 2)
        self.assertEqual(b[1], Bool(0))
        self.assertEqual(b[0], Bool(1))

        c = a[0:3]
        self.assertIsInstance(c, BoolArray)
        assert isinstance(c, BoolArray)
        self.assertEqual(len(c), 4)
        self.assertEqual(c[3], Bool(1))
        self.assertEqual(c[2], Bool(1))
        self.assertEqual(c[1], Bool(0))
        self.assertEqual(c[0], Bool(1))

    def test_bool_array_var(self):
        a = ArrayVariable(BoolArray("101011"))
        self.assertEqual(a.evaluate(), BoolArray("101011"))
        self.assertRaises(errors.StaplValueError, lambda: a.assign(IntArray([1, 2, 3, 4, 5, 6])))
        a.assign(BoolArray("110011"))
        self.assertEqual(a.evaluate(), BoolArray("110011"))
        a = ArrayVariable(BoolArray("000000"))
        a.assign(slice(4, 3), BoolArray("11"))
        self.assertEqual(a.evaluate(), BoolArray("011000"))
        a.assign(slice(1, 0), BoolArray("10"))
        self.assertEqual(a.evaluate(), BoolArray("011010"))
        a.assign(3, Bool(0))
        self.assertEqual(a.evaluate(), BoolArray("010010"))
        a.assign(slice(0, 2), BoolArray("001"))
        self.assertEqual(a.evaluate(), BoolArray("010100"))
        a.assign(BoolArray("01"))
        self.assertEqual(a.evaluate(), BoolArray("000001"))
        a.assign(BoolArray("1111111111110"))
        self.assertEqual(a.evaluate(), BoolArray("111110"))
        a.assign(slice(5, 4), BoolArray("100"))
        self.assertEqual(a.evaluate(), BoolArray("001110"))
        a.assign(slice(2, 0), BoolArray("1"))
        self.assertEqual(a.evaluate(), BoolArray("001001"))

if __name__ == '__main__':
    unittest.main()

